/*
 * A theory of arrays
 * (setq tab-width 2)
 */

THEORY Array

DEFINITIONS

DATATYPES
/*
 * datatype elem  = elem of nat;
 *
 * datatype array = aempty
 *             | array of nat * elem * array
 *
 * datatype array = elem list;
 */
datatype elem  = elem of nat;
         array = elem list

END

PREDICATES

create : nat --> elem --> array --> prop <=
       \Size:nat. \Initval:elem. \Res:array.
            cases (Size, Res) of
              (0, [] : array)
            | (S'+1, [Initval | Rest]) => 
                     create S' Initval Rest
            end
end;

at : array --> nat --> elem --> prop <=
   \A:array. \N:nat. \E:elem.
       cases A of
       [Head | Tail] =>
             cases N of
                 0    => Head = E
               | N'+1 => at Tail N' E
            end
       end
end;

set : array --> nat --> elem --> array --> prop <=
    \A:array. \N:nat. \E:elem. \A':array.
        cases (A, A') of
        ([H | T], [H' | T']) =>
            cases N of
              0    => (E=H' and T=T')
            | N'+1 => (H=H' and set T N' E T')
            end
        end
end;

size : array --> nat --> prop <=
     \A:array. \N:nat.
          cases A of
            [Head | Tail] =>
                  if   Tail = [] : array
                  then N = 1
                  else cases N of
                        N'+1 => size Tail N'
                       end
          end
end;

extend : array --> elem --> array --> prop <=
       \A:array. \E:elem. \A':array.
            cases A of
              []:array => [E] = A'
            | [H | T]  => 
                 cases A' of
                 [H | T'] => extend T E T'
                 end
            end
end;

contains : array --> elem --> prop <=
         \A:array. \E:elem.
         cases A of
         [Head | Tail] =>
               if   Head = E
               then tt
               else contains Tail E
         end
end;


LET
reverse : array --> array --> prop =
         \A1:array. \A2:array.
         impl A1 []:array A2
WHERE
impl : array --> array --> array --> prop <=
         \A1:array. \A2:array. \Res:array.
         cases A1 of
           []:array => A2 = Res
         | [H | T]  => exists A2':array.
                       A2' = [H | A2] and
                       impl T A2' Res
         end         
end
END;

LET
set_from : array --> nat --> array --> array --> prop <=
         \A1:array. \N:nat. \A2:array. \A3:array. 
         cases (N, A1, A3) of
           (0, X, Y) => impl A1 A2 A3
         | (N'+1, [H | T1], [H | T3]) => 
                     set_from T1 N' A2 T3
         end
WHERE
impl : array --> array --> array --> prop <=
     \A1:array. \A2:array. \A3:array.
     cases A2 of
       []:array  => A1 = A3
     | [H2 | T2] =>
           cases (A1, A3) of
           ([H1 | T1], [H2 | T3]) =>
                impl T1 T2 T3
           end
     end
end
END;

split : array --> nat --> array --> array --> prop <=
      \A1:array. \N:nat. \A2:array. \A3:array.
      cases (A1, N, A2) of
        (X, 0, []:array) => A1 = A3
      | ([H | T1], N'+1, [H | T2]) =>
            split T1 N' T2 A3
      end
end;

LET
split_rev : array --> nat --> array --> array --> prop =
      \A1:array. \N:nat. \A2:array. \A3:array.
      impl A1 N []:array A2 A3
WHERE
impl : array --> nat --> array --> array --> array --> prop <=
     \A:array. \N:nat. \Acc:array. \Top:array. \Rest:array.
      cases (A, N) of
        (X, 0) => Top = Acc and A = Rest
      | ([H | T], N'+1) =>
            impl T N' [H | Acc] Top Rest
      end
end
END;

remove : array --> nat --> array --> prop <=
       \A1:array. \N:nat. \A2:array.
       cases (A1, N) of
         ([H1 | T1 ], 0)    => A2 = T1
       | ([H1 | T1 ], N'+1) => 
           cases A2 of
	   [H1 | T2] => remove T1 N' T2
	   end
       end
END;

/* Temporary place-fillment formula. */
fill : nat --> prop =
     \N:nat.
        N = 0
end
END

/* Accessor functions. */
SML FROM "ArrayAccess.sml"


SML {*
fun isSetFromL pos seq = 
    Array.is_set_from_pred (Sequent.get_formula_lhs(seq,pos));;

fun isSetFromR pos seq = 
    Array.is_set_from_pred (Sequent.get_formula_rhs(seq,pos));;
*}


/* Tactics and scripts for this theory */
SML FROM "tactics.sml"

/* with elem=nat*nat this is interesting
THEOREM ex
declare N:nat in
|- exists N':nat. contains [(0,1), (2, 3), (7,4)] (2, N') and N' = 3
END
*/


THEOREM ex22
declare E1:elem, E2:elem, E3:elem, X:elem, A:array, A2:array, Np:nat in 
/* Example to describe split to the left.*/
fill 1, remove [E1, E2, E3] 1 A, fill 2 |-
 fill 3, A = [E1, E3], fill 4
/*
PROOF {*
      Array_remove_l 2;;
      Eq_r 2;;
*}*/
END

THEOREM ex
declare E1:elem, E2:elem, E3:elem, X:elem, A:array, A2:array, Np:nat in 
/* Example to describe remove to the right.*/
fill 1, fill 2 |-
 fill 3, 
exists AA:array.
       remove [E1, E2, E3] 1 AA and
       AA = [E1, E3]
, fill 4
/*PROOF {*
        Exists_r NONE 2;;
        And_r 2;;
        Array_remove_r 2;;
        Eq_r 2;;      
*}*/
END

THEOREM ex21
declare E1:elem, E2:elem, E3:elem, X:elem, A:array, A2:array, Np:nat in 
/* Example to describe remove to the right.*/
fill 1, fill 2 |-
 fill 3, 
exists AA:array. exists BB:array.
       split_rev [E1, E2, E3, X] 2 AA BB and
       AA = [E2, E1] and
       BB = [E3, X]
, fill 4
/*PROOF {*
        Exists_r NONE 2;;
        Exists_r NONE 2;;
        And_r 2;;
        Array_split_rev_r 2;;
        And_r 2;;
        Eq_r 2;;
        Eq_r 2;;      
*}*/
END


THEOREM ex20
declare E1:elem, E2:elem, E3:elem, X:elem, A:array, A2:array, Np:nat in 
/* Example to describe split_rev to the left.*/
fill 1, split_rev [E1, E2, E3] 0 A A2, fill 2 |-
 fill 3, A = [E2, E1], A2 = [E3], fill 4
/*
PROOF {*
      Array_split_rev_l 2;;
      Eq_r 2;;
*}*/
END

THEOREM ex19
declare E1:elem, E2:elem, E3:elem, X:elem, A:array, A2:array, Np:nat in 
/* Example to describe split to the right.*/
fill 1, fill 2 |-
 fill 3, 
exists AA:array. exists BB:array.
       split [E1, E2, E3, X] 2 AA BB and
       AA = [E2, E1] and
       BB = [E3]
, fill 4
/*PROOF {*
        Exists_r NONE 2;;
        Exists_r NONE 2;;
        And_r 2;;
        Array_split_rev_r 2;;
        And_r 2;;
        Eq_r 2;;
        Eq_r 2;;      
*}*/
END


THEOREM ex18
declare E1:elem, E2:elem, E3:elem, X:elem, A:array, A2:array, Np:nat in 
/* Example to describe split to the left.*/
fill 1, split [E1, E2, E3] 2 A A2, fill 2 |-
 fill 3, A = [E1, E2], A2 = [E3], fill 4
/*
PROOF {*
      Array_split_l 2;;
      Eq_r 2;;
*}*/
END

THEOREM ex17
declare E1:elem, E2:elem, E3:elem, E4:elem, E5:elem, E6:elem, X:elem, A:array, A2:array, Np:nat in 
/* Example to describe set_from to the right.*/
fill 1, fill 2 |-
 fill 3, exists A':array. set_from [E1, E2, E3, E4] 1 [E5, E6] A' and A' = [E1, E5, E6, E4], fill 4
/*PROOF {*
      Array_create_l 2;;
      Array_at_r 2;;
*}*/
END

THEOREM ex16
declare E1:elem, E2:elem, E3:elem, E4:elem, E5:elem, E6:elem, X:elem, A:array, A2:array, Np:nat in 
/* Example to describe set_from to the left.*/
fill 1, set_from [E1, E2, E3, E4] 1 [E5, E6] A, fill 2 |-
 fill 3, A = [E1, E5, E6, E4], fill 4
/*PROOF {*
      Array_create_l 2;;
      Array_at_r 2;;
*}*/
END

THEOREM ex15
declare E1:elem, E2:elem, E3:elem, X:elem, A:array, A2:array, Np:nat in 
/* Example to describe reverse to the right.*/
fill 1, fill 2 |-
 fill 3, exists A3:array. reverse [E1, E2, E3] A3 and A3 = [E3, E2, E1], fill 4
/*PROOF {*
        Exists_r NONE 2;;
        And_r 2;;
        Array_reverse_r 2;;
        Eq_r 2;;
*}*/
END

THEOREM ex14
declare E1:elem, E2:elem, E3:elem, X:elem, A:array, A2:array, Np:nat in 
/* Example to describe reverse to the left.*/
fill 1, reverse [E1, E2, E3] A, fill 2 |-
 fill 3, A = [E3, E2, E1], fill 4
/*PROOF {*
      Array_create_l 2;;
      Array_at_r 2;;
*}*/
END

THEOREM ex13
declare E1:elem, E2:elem, E3:elem, X:elem, A:array, A2:array, Np:nat in 
/* Example to describe create to the right.*/
fill 1, fill 2 |-
 fill 3, exists A1:array. create 2 E1 A1 and at A1 1 E1, fill 4
/*PROOF {*
        Exists_r NONE 2;;
        And_r 2;;
        Array_create_r 2;;
        Array_at_r 2;;
*}*/
END

THEOREM ex12
declare E1:elem, E2:elem, E3:elem, X:elem, A:array, A2:array, Np:nat in 
/* Example to describe create to the left.*/
fill 1, create 2 E1 A, fill 2 |-
 fill 3, at A 1 E1, fill 4
/*PROOF {*
      Array_create_l 2;;
      Array_at_r 2;;
*}*/
END

THEOREM ex11
declare E1:elem, E2:elem, E3:elem, X:elem, A:array, A2:array, Np:nat in 
/* Example to describe contains to the right.*/
fill 1, fill 2 |-
 fill 3, exists X':elem. if contains [elem(0),elem(1)] X' then X'=elem(1) else not X'=elem(1), fill 4
/*PROOF {*
      Array_contains_r 2;;
*}*/
END

THEOREM ex10
declare E1:elem, E2:elem, E3:elem, X:elem, A:array, A2:array, Np:nat in 
/* Example to describe contains to the right.*/
fill 1, fill 2 |-
 fill 3, contains [elem(0),elem(1)] elem(0), fill 4
/*PROOF {*
      Array_contains_r 2;;
*}*/
END


THEOREM ex9
declare E1:elem, E2:elem, E3:elem,  E4:elem, X:elem, A:array, A2:array, Np:nat in 
/* Example to describe contains to the left.*/
fill 1, contains [elem(0), elem(1), elem(2)] X, fill 2 |-
 fill 3, X = E2, fill 4
/*
PROOF {*
      Array_contains_l 2;;
      Eq_r 2
*}*/
END


THEOREM ex8
declare E1:elem, E2:elem, E3:elem, X:elem, A:array, A2:array, Np:nat in 
/* Example to describe extendArray to the right.*/
fill 1, fill 2 |-
 fill 3, 
 extend
    [E1, E2]
    E3 
    [E1, E2, E3],
 fill 4
/*PROOF {*
      Array_extend_r 2
*}*/
END

THEOREM ex7
declare E1:elem, E2:elem, E3:elem, X:elem, A:array, A2:array, Np:nat in 
/* Example to describe extendArray to the left.*/
fill 1, extend [E1, E2] E3 A2, fill 2 |-
 fill 3, at A2 2 E3, fill 4
/*PROOF {*
      Array_extend_l 2;;
      Array_at_r 2
*}*/
END


THEOREM ex6
declare E1:elem, E2:elem, E3:elem, X:elem, A:array, A2:array, Np:nat in 
/* Example to describe size to the left. */
fill 1, size [E1, E2] Np, fill 2 |-
 fill 3, Np=2, fill 4
/*PROOF {*
      Array_size_l 2;;
      Eq_r 2;;
*}*/
END


THEOREM ex5
declare E1:elem, E2:elem, E3:elem, X:elem, A:array, A2:array, Np:nat in 
/* Example to describe size to the right.*/
fill 1, fill 2 |-
 fill 3, size [E1, E2]  2, fill 4
/*PROOF {*
      Array_size_r 2
*}*/
END

THEOREM ex4
declare E1:elem, E2:elem, E3:elem, X:elem, A:array, A2:array, Np:nat in 
/* Example to describe set on the right.*/
fill 1, fill 2 |-
 fill 3, exists A3:array. set [E1, E3 | A] 1 E2 A3, fill 4
/*PROOF {*
      Exists_r NONE 2;;
      Array_set_r 2
*}*/
END


THEOREM ex3
declare E1:elem, E2:elem, E3:elem, X:elem, A:array, A2:array, Np:nat in 
/* Example to describe set on the left.*/
fill 1, set [E1, E3 | A] 1 E2 A2, fill 2 |-
 fill 3, at A2 1 E2, fill 4
/*PROOF {*
      Array_set_l 2;;
      Array_at_r 2;;
*}*/
END


THEOREM ex2
declare E1:elem, E2:elem, E3:elem, X:elem, A:array, A2:array, Np:nat in 
/* Example to describe at to the right.*/
fill 1, fill 2 |-
 fill 3, at [E1, E2, E3 | A] 1 E2, fill 4
/*PROOF {*
      Array_at_r 2;;
*}*/
END


THEOREM ex1
declare E1:elem, E2:elem, E3:elem, X:elem, A:array, A2:array, Np:nat in 
/* Example to describe at to the left.*/
fill 1, at [E1, E2, E3] 1 X, fill 2 |-
 fill 3, X = E2, fill 4
/*
PROOF {*
      Array_at_l 2;;
      Eq_r 2
*}*/
END

END /* theory */
