/*
 * A theory of arrays
 * (setq tab-width 2)
 */

THEORY Array2

DEFINITIONS

DATATYPES
datatype elem  = elem of nat;

datatype array = aempty
							 | array of nat * elem * array
END

PREDICATES

at : array --> nat --> elem --> prop <=
   \A:array. \N:nat. \E:elem.
	     cases A of
	     array(Idx, Top, Rest) =>
			 if N = Idx then E=Top
									else at Rest N E
	     end
end;

LET
wfArray: array --> prop =
				 \A:array.
				 cases A of
							 array(0, E, Ap) => wfArrayImpl Ap 0
				 |     aempty()
				 end
WHERE
				wfArrayImpl: array --> nat --> prop <=
				\A:array. \N:nat.
				  cases A of
				  			array(Idx, E, Ap) => N+1 = Idx and
				  													 wfArrayImpl Ap Idx
				  |     aempty()
				  end
				end
END;

LET
extendArray : array --> elem --> array --> prop =
   \A1:array. \E:elem. \A2:array.
				wfArray A1 and
	      cases (A1, A2) of
	        (array(Idx, Top, Rest1), array(Idx, Top, Rest2)) =>
					  extendArrayImpl Rest1 E Rest2 Idx
				| (aempty(), array(0, E, aempty()))
				end
WHERE
  extendArrayImpl : array --> elem --> array --> nat --> prop <=
     \A1:array. \E:elem. \A2:array. \PIdx:nat.  
  	      cases (A1, A2) of
  	        (array(Idx, Top, Rest1), array(Idx, Top, Rest2)) =>
  					  extendArrayImpl Rest1 E Rest2 Idx
  				| (aempty(), array(PIdx+1, E, aempty()))
  				end
  end
END;

set : array --> nat --> elem --> array --> prop <=
		\A1:array. \N:nat. \E:elem. \A2:array.
		cases (A1, A2) of
		  (array(Idx, Top1, Rest1), array(Idx, Top2, Rest2)) =>
									if N=Idx then (E=Top2 and Rest1 = Rest2)
													 else (Top1=Top2 and set Rest1 N E Rest2)
		end
end;

LET
size : array --> nat --> prop <=
		 \A:array. \Size:nat.
		 wfArray A and
		 cases A of
					 array(Curr, E, Ap) => sizeImpl A Curr Size
		 |     aempty() => Size = 0
		 end
WHERE
				sizeImpl: array --> nat --> nat --> prop <=
				\A:array. \Last:nat.\Size:nat.
				  cases A of
				  			array(Curr, E, Ap) => sizeImpl Ap Curr Size
				  |     aempty() => Last+1 = Size
				  end
				end
END;

/* Temporary place-fillment formula. */
fill : nat --> prop =
		 \N:nat.
		 exists A:array. exists E:elem.
										 set A N E A
end
END

/* Predicate accessors defined in Standard ML * /
SML FROM "Array2Access.sml"
/ * Tactics and scripts for this theory */
SML FROM "tactics.sml"

THEOREM ex
declare E1:elem, E2:elem, E3:elem, X:elem, A:array, A2:array, Np:nat in 
/* Example to describe extendArray to the right.*/
fill 1, fill 2 |-
 fill 3, 
 extendArray
		aempty()
		E3 
		array(0, E3, aempty()), 
 fill 4
/*PROOF {*
			Array_extend_r 2;;
*}*/
END


THEOREM ex10
declare E1:elem, E2:elem, E3:elem, X:elem, A:array, A2:array, Np:nat in 
/* Example to describe extendArray to the right.*/
fill 1, fill 2 |-
 fill 3, 
 extendArray
		array(0, E1, array(1, E2, aempty())) 
		E3 
		array(0, E1, array(1, E2, array(2, E3, aempty()))), 
 fill 4
PROOF {*
			Array_extend_r 2
*}
END

THEOREM ex9
declare E1:elem, E2:elem, E3:elem, X:elem, A:array, A2:array, Np:nat in 
/* Example to describe extendArray to the left.*/
fill 1, extendArray array(0, E1, array(1, E2, aempty())) E3 A2, fill 2 |-
 fill 3, at A2 2 E3, fill 4
PROOF {*
			Array_extend_l 2;;
			Array_at_r 2
*}
END

THEOREM ex8
declare E1:elem, E2:elem, E3:elem, X:elem, A:array, A2:array, Np:nat in 
/* Example to describe size to the left. */
fill 1, size array(0, E1, array(1, E2, aempty())) Np, fill 2 |-
 fill 3, Np=2, fill 4
PROOF {*
			Array_size_l 2;;
			Eq_r 2;;
*}
END


THEOREM ex7
declare E1:elem, E2:elem, E3:elem, X:elem, A:array, A2:array, Np:nat in 
/* Example to describe size to the right.*/
fill 1, fill 2 |-
 fill 3, size array(0, E1, array(1, E2, aempty())) 2, fill 4
PROOF {*
			Array_size_r 2
*}
END


THEOREM ex6
declare E1:elem, E2:elem, E3:elem, X:elem, A:array, A2:array, Np:nat in 
/* Example to describe wfArray to the right.*/
fill 1, fill 2 |-
 fill 3, wfArray array(0, E1, array(1, E2, aempty())), fill 4
PROOF {*
			Array_wf_r 2
*}
END


THEOREM ex5
declare E1:elem, E2:elem, E3:elem, X:elem, A:array, A2:array, Np:nat in 
/* Example to describe wfArray to the left.*/
fill 1, wfArray array(0, E1, array(1, E2, aempty())), fill 2 |-
 fill 3, tt, fill 4
PROOF {*
			Array_wf_l 2;;
			True_r 2;;
*}
END


THEOREM ex4
declare E1:elem, E2:elem, E3:elem, X:elem, A:array, A2:array, Np:nat in 
/* Example to describe set on the right.*/
fill 1, fill 2 |-
 fill 3, set array(0, E1, array(1, E3, A)) 1 E2 
						 array(0, E1, array(1, E2, A)), fill 4
PROOF {*
			Array_set_r 2
*}
END


THEOREM ex3
declare E1:elem, E2:elem, E3:elem, X:elem, A:array, A2:array, Np:nat in 
/* Example to describe set on the left.*/
fill 1, set array(0, E1, array(1, E3, A)) 1 E2 A2, fill 2 |-
 fill 3, at A2 1 E2, fill 4
PROOF {*
			Array_set_l 2;;
			Array_at_r 2;;
*}
END


THEOREM ex2
declare E1:elem, E2:elem, E3:elem, X:elem, A:array, A2:array, Np:nat in 
/* Example to describe at to the right.*/
fill 1, fill 2 |-
 fill 3, at array(0, E1, array(1, E2, array(2, E3, A))) 1 E2, fill 4
PROOF {*
			Array_at_r 2
*}
END


THEOREM ex1
declare E1:elem, E2:elem, E3:elem, X:elem, A:array, A2:array, Np:nat in 
/* Example to describe at to the left.*/
fill 1, at array(0, E1, array(1, E2, array(2, E3, A))) 1 X, fill 2 |-
 fill 3, X = E2, fill 4
PROOF {*
			Array_at_l 2;;
			Eq_r 2;;
*}
END

END /* theory */
