(*
 * Tactics-file for assisting the user in the task of handling 
 * of Arrays.
 *
 * @todo REALLY shold insert check into the t_while' tactics to make
 * sure that they act on the correct formula. Could be a source of unintended 
 * interesting behaviur otherwise.
 *)


(*****************************************
     Utility tactics
******************************************)
(* Tactics for applying formulas to the end of sequents
 * to the left and the right. Copied from the reference
 * manual.
 *)
val apply_to_end_r =
    t_pretactic
    (fn seq => SOME(length(Sequent.get_rhs seq)))
val apply_to_end_l =
    t_pretactic
    (fn seq => SOME(length(Sequent.get_lhs seq)))
	    
    
(*****************************************
     Tactics handling at-formula
******************************************)
(* Handles one unfolding of an at-formula on the right.
*)
fun array_at_r_step pos =
    t_compose_l
    [unfold_r pos,
     t_succeed (unfold_r pos),
     apply_r pos,
     apply_r pos,
     apply_r pos,
     cases_r pos,
     (t_compose
	  (if_then_else_r pos)
	  [(t_orelse (*Handle true-part*)
		(apply_to_end_l eq_flat_elim_l) (* Wrong place? *)
		(eq_r pos)                      (* Right place! *)
		),
	   (t_orelse (*Handle false-part*)
		(apply_to_end_r eq_flat_elim_r) (* Wrong place? *)
		(apply_to_end_r eq_r)           (* stop searching. *)
		)
	   ]
	  )
     ]
(* Succeeds if the at-formula on the right is true.
*)
fun array_at_r pos = t_while'(array_at_r_step pos)

(* Unfolds and splits the cases of an t-formula on the left
*)
fun array_at_l_step pos =
    t_compose_l
	[unfold_l pos,
	 t_succeed (unfold_l pos),
	 apply_l pos,
	 apply_l pos,
	 apply_l pos,
	 cases_l pos,
	 apply_to_end_l eq_subst,
	 apply_to_end_l eq_subst,
	 t_compose
	     (if_then_else_l pos)
	     [(*Handle true-part*)
	      (t_orelse
		   (apply_to_end_l eq_flat_elim_l)  (* Wrong place? *)
		   (t_compose2 (apply_to_end_l eq_l)(* Otherwise, the equality holds...*)
			       (eq_subst pos)       (* ...so we have our value *)
			       )           
		   ),
	      (*Handle false-part*)
	      (t_orelse
		   (apply_to_end_r eq_flat_elim_r) (* Wrong place? *)
		   (apply_to_end_r eq_r)           (* otherwize, we can stop searching.*)
		   )
	      ]
	 ]
(* Do as many steps as possible of the above tactic, at least one.
*)
fun array_at_l pos = t_while'(array_at_l_step pos)


(*****************************************
     Tactics handling set-formula
******************************************)
fun array_set_l_step pos =
    t_compose_l
	[unfold_l pos,
	 t_succeed (unfold_l pos),
	 apply_l pos,
	 apply_l pos,
	 apply_l pos,
	 apply_l pos,
	 cases_l pos,
	 (apply_to_end_l eq_flat_decomp_l),
	 (apply_to_end_l eq_subst),
	 (apply_to_end_l eq_subst),
	 (apply_to_end_l eq_subst),
	 (t_compose
	      (if_then_else_l pos)
	      [(* Handle true-part *)
	       (t_orelse
		    (apply_to_end_l eq_flat_elim_l) (*Wrong place?*)
		    (t_compose_l  (* Substitute the vals *)
			 [and_l pos,
			  eq_subst pos,
			  eq_subst pos,
			  (apply_to_end_l eq_l)
			  ]
			 )
		    ),
	       (* Handle false-part*)
	       (t_orelse
		    (apply_to_end_r eq_r)
		    (t_compose_l
			 [and_l pos,
			  eq_subst pos,
			  (apply_to_end_r eq_flat_elim_r)
			  ]
			 )
		    ) (*end, orelse*)
	       ] (*end, compose *)
	  )
	 ]
(* Do as many steps as possible of the above tactic, at least one.
*)
fun array_set_l pos = t_while'(array_set_l_step pos)

(* Handles on unfolding of an set-formula on the right.
*)
fun array_set_r_step pos =
    t_compose_l
	[unfold_r pos,
	 t_succeed (unfold_r pos),
	 apply_r pos,
	 apply_r pos,
	 apply_r pos,
	 apply_r pos,
	 cases_r pos,
	 (t_compose
	      (if_then_else_r pos)
	      [(*Handle true-part*)
	       (t_orelse
		    (apply_to_end_l eq_flat_elim_l)
		    (t_compose
			 (and_r pos)
			 [ (* Finish off the equalities. *)
			  (eq_r pos),
			  (eq_r pos)
			  ]
		     )
	       ),
	       (*Handle false-part*)
	       (t_orelse 
		    (apply_to_end_r eq_r)
		    (t_compose
			 (and_r pos)
			 [ (*Check for equal tops *)
			  (eq_r pos),
			  (* Remove false formula *)
			  (apply_to_end_r eq_flat_elim_r)
			  ]
			 )
		    )
	       ] (*end, compose for *)
	  )
	 ]
(* Do as many steps as possible of the above tactic, at least one.
*)
fun array_set_r pos = t_while'(array_set_r_step pos)


(*****************************************
     Tactics handling wfArray-formula
******************************************)
fun array_wfImpl_l_step pos =
    t_compose_l
	[unfold_l pos,
	 apply_l pos,
	 apply_l pos,
	 cases_l pos,
	 (t_orelse
	      (true_l pos)
	      (t_compose_l
		   [(apply_to_end_l eq_subst),
		    (apply_to_end_l eq_subst),
		    and_l pos,
		    eq_l pos
		    ]
	       )
	      )
	 ]
fun array_wf_l_step     pos =
    t_compose_l 
	[t_succeed (unfold_l pos),
	 apply_l pos,
	 cases_l pos,
	 (t_orelse
	      (true_l pos)
	      (t_compose_l
		   [(apply_to_end_l eq_subst),
		    (apply_to_end_l eq_l)
		    ]
	       )
	  )
	 ]
fun array_wf_l          pos =
    t_compose_l
	[array_wf_l_step pos,
	 (t_while (array_wfImpl_l_step pos))
	 ]

fun array_wfImpl_r_step pos =
    t_compose_l
	[unfold_r pos,
	 apply_r pos,
	 apply_r pos,
	 cases_r pos,
	 t_orelse
	     (true_r pos)
	     (t_compose
		  (and_r pos)
		  [(eq_r pos),
		   t_skip
		   ]
		  )
	 ]
fun array_wf_r_step     pos =
    t_compose_l
	[t_succeed (unfold_r pos),
	 apply_r pos,
	 cases_r pos,
	 t_succeed (true_r pos)
	 ]
fun array_wf_r pos =
    t_compose_l
	[array_wf_r_step pos,
	 (t_while (array_wfImpl_r_step pos))
	 ]


(*****************************************
     Tactics handling size-formula
******************************************)
fun array_sizeImpl_r_step pos =
    t_compose_l
	[unfold_r pos,
	 apply_r pos,
	 apply_r pos,
	 apply_r pos,
	 cases_r pos,
	 t_succeed (eq_r pos)
	 ]
fun array_size_r_step     pos =
    t_compose_l
	[unfold_r pos,
	 unfold_r pos,
	 apply_r pos,
	 apply_r pos,
	 t_compose
	     (and_r pos)
	     [array_wf_r pos,
	      t_compose_l 
		  [cases_r pos,
		   t_succeed (eq_r pos)
		   ]
	      ]
	 ]
fun array_size_r          pos =
    t_compose_l
	[array_size_r_step pos,
	 (t_while (array_sizeImpl_r_step pos))
	 ]

fun array_sizeImpl_l_step pos =
    t_compose_l
	[unfold_l pos,
	 apply_l pos,
	 apply_l pos,
	 apply_l pos,
	 cases_l pos,
	 t_orelse
	     (eq_subst pos)
	     (t_compose_l
		  [(apply_to_end_l eq_subst),
		   (apply_to_end_l eq_subst)
		   ]
		  )
	 ]
fun array_size_l_step     pos =
    t_compose_l
	[unfold_l pos,
	 unfold_l pos,
	 apply_l pos,
	 apply_l pos,
	 and_l pos,
	 (* Rather ugly hack, since cases_l, which is
	  * used in arra_wf_l, affects subsequent cases in
	  * the formula.
	  *)
	 move_to_end_l pos,
	 (apply_to_end_l array_wf_l),
	 cases_l pos,
	 t_orelse
	     (eq_subst pos)
	     (t_compose_l 
		  [(apply_to_end_l eq_subst),
		   (apply_to_end_l eq_subst)
		   ]
		  )
	 ]
fun array_size_l          pos =
    t_compose_l
	[array_size_l_step pos,
	 (t_while (array_sizeImpl_l_step pos))
	 ]

(*****************************************
     Tactics handling extendArray-formula
******************************************)
fun array_extendImpl_l_step pos =
    t_compose_l
	[unfold_l pos,
	 apply_l pos,
	 apply_l pos,
	 apply_l pos,
	 apply_l pos,
	 t_compose
	     (cases_l pos)
	     [(t_orelse
		   (apply_to_end_l eq_flat_elim_l)
		   (t_compose_l
			[apply_to_end_l eq_flat_decomp_l,
			 apply_to_end_l eq_subst,
			 apply_to_end_l eq_subst,
			 apply_to_end_l eq_subst
			 ])
	       ),
	      (t_orelse
		   (apply_to_end_l eq_flat_elim_l)
		   (t_compose_l
			[true_l pos,
			 apply_to_end_l eq_l
			 ])
	       )
	      ]
	 ]
fun array_extend_l_step     pos =
    t_compose_l
	[unfold_l pos,
	 apply_l pos,
	 apply_l pos,
	 apply_l pos,
	 and_l pos,
	 move_to_end_l pos, (*UGLY HACK, should be moved into cases_l really*)
	 (apply_to_end_l array_wf_l),
	 t_compose
	     (cases_l pos)
	     [(t_orelse
		   (apply_to_end_l eq_flat_elim_l)
		   (t_compose_l
			[apply_to_end_l eq_flat_decomp_l,
			 apply_to_end_l eq_subst,
			 apply_to_end_l eq_subst,
			 apply_to_end_l eq_subst
			 ])
	       ),
	      (t_orelse
		   (apply_to_end_l eq_flat_elim_l)
		   (t_compose_l
			[true_l pos,
			 apply_to_end_l eq_l
			 ])
	       )
	      ]
	 ]
fun array_extend_l          pos =
    t_compose_l
	[array_extend_l_step pos,
	 (t_while (array_extendImpl_l_step pos))
	 ]


fun array_extendImpl_r_step pos =
    t_compose_l
	[unfold_r pos,
	 apply_r pos,
	 apply_r pos,
	 apply_r pos,
	 apply_r pos,
	 cases_r pos,
	 t_succeed (true_r pos)
	 ]
fun array_extend_r_step     pos =
    t_compose_l
	[unfold_r pos,
	 apply_r pos,
	 apply_r pos,
	 apply_r pos,
	 t_compose
	     (and_r pos)
	     [array_wf_r pos,
	      t_compose_l 
		  [cases_r pos,
		   t_succeed (true_r pos)
		   ]
	      ]
	 ]
fun array_extend_r          pos =
    t_compose_l
	[array_extend_r_step pos,
	 (t_while (array_extendImpl_r_step pos))
	 ]



(*****************************************
     Commands to ease the use of the
     above tactics in the interactive 
     assistant.
******************************************)
fun Array_at_r     pos = by(array_at_r pos)
fun Array_at_l     pos = by(array_at_l pos)
fun Array_wf_l     pos = by(array_wf_l pos)
fun Array_wf_r     pos = by(array_wf_r pos)
fun Array_set_l    pos = by(array_set_l pos)
fun Array_set_r    pos = by(array_set_r pos)
fun Array_size_r   pos = by(array_size_r pos)
fun Array_size_l   pos = by(array_size_l pos)
fun Array_extend_l pos = by(array_extend_l pos)
fun Array_extend_r pos = by(array_extend_r pos)
