(*
 * Tactics-file for assisting the user in the task of handling 
 * of Arrays.
 *
 * @todo REALLY shold insert check into the t_while' tactics to make
 * sure that they act on the correct formula. Could be a source of unintended 
 * interesting behaviur otherwise.
 *)


(*****************************************
     Utility tactics
******************************************)
(* Tactics for applying formulas to the end of sequents
 * to the left and the right. Copied from the reference
 * manual.
 *)
val apply_to_end_r =
    t_pretactic
	(fn seq => SOME(length(Sequent.get_rhs seq)))
val apply_to_end_l =
    t_pretactic
	(fn seq => SOME(length(Sequent.get_lhs seq)))
(* Length of assumptions to the left *)
fun len_assump_l s = List.length(Sequent.get_lhs s)
fun some_len_assump_l s = SOME(len_assump_l s)
(* Length of assumptions to the right *)
fun len_assump_r s = List.length(Sequent.get_rhs s)
fun some_len_assump_r s = SOME(len_assump_r s)
			  
fun handleif_l testl testr pos bodyt bodyf =
    t_compose
	(if_then_else_l pos)
	[t_compose_l
	     ([apply_to_end_l testl
	       ] @ bodyt),
	     t_compose_l
		 ([apply_to_end_r testr
		   ] @ bodyf)
		 ]
fun handleif_r testl testr pos bodyt bodyf =
    t_compose
	(if_then_else_r pos)
	[t_compose_l
	     ([apply_to_end_l testl
	       ] @ bodyt),
	     t_compose_l
		 ([apply_to_end_r testr
		   ] @ bodyf)
		 ]

(* Should we move/insert exteq_[lr] before the loop in both 
 * try_eq_[lr] respectively?
 *)

fun handle_eq_l pos =
    (t_orelse_l
	 [eq_l pos,
	  eq_subst pos,
	  eq_flat_elim_l pos,
	  eq_flat_decomp_l pos
	  ]
	 )
   
fun try_eq_l pos =
    t_bool (fn seq => Mmc.is_eq(Sequent.get_formula_lhs(seq,pos)))
    (t_compose_l
	[move_to_end_l pos,
	 t_pretactic some_len_assump_l
	 (fn from_pos =>
	     (t_while(t_to_end_l from_pos handle_eq_l))
	     (* This code would check that indeed, we have checked all of the equality.
	      * t_bool (fn seq => ((len_assump_l seq) = from_pos-1))
		       t_skip
		       t_fail
	      *)
	     )
	 ])
    t_fail

fun handle_eq_r pos =
    (t_orelse_l
	 [exteq_r pos,
	  eq_flat_elim_r pos,
	  eq_flat_decomp_r pos
	  ]
	 )
   
fun try_eq_r pos =
    t_bool (fn seq => Mmc.is_eq(Sequent.get_formula_rhs(seq,pos)))
    (t_compose_l
	[move_to_end_r pos,
	 t_pretactic some_len_assump_r
	 (fn from_pos =>
	     (t_while(t_to_end_r from_pos handle_eq_r))
	     (* This code would check that indeed, we have checked all of the equality.
	      * t_bool (fn seq => ((len_assump_r seq) = from_pos-1))
		       t_skip
		       t_fail
	      *)
	     )
	 ])
    t_fail

fun is_eq_r pos =
    (t_orelse_l
	 [eq_r pos,
	  eq_flat_elim_r pos,
	  eq_flat_decomp_r pos
	  ]
	 )
   
fun check_eq_r pos =
    t_bool (fn seq => Mmc.is_eq(Sequent.get_formula_rhs(seq,pos)))
    (t_compose_l
	[move_to_end_r pos,
	 t_pretactic some_len_assump_r
	 (fn from_pos =>
	     (t_compose_l
		  [t_while(t_to_end_r from_pos is_eq_r)(*,
		   (* This code would check that indeed, we have checked all of the equality.
		    * t_bool (fn seq => ((len_assump_r seq) = from_pos-1))
			     t_skip
			     t_fail
		    *)
		   (move_segment_r from_pos len_assump_r pos)*)
		   ])
	     )
	 
	 ])
    t_fail


val handleifeq_l = handleif_l handle_eq_l handle_eq_r
val handleifeq_r = handleif_r handle_eq_l handle_eq_r



(*****************************************
     Tactics handling create-formula
******************************************)
fun array_create_l pos = 
    t_fix ()
    (fn () =>
	(fn recurse =>
	    t_compose_l
		[unfold_l pos,
		 t_succeed (unfold_l pos),
		 apply_l pos,
		 apply_l pos,
		 apply_l pos,
		 cases_l pos,
		 t_orelse_l
		     [true_l pos,
		      recurse()
		      ]
		 ]))
    
fun array_create_r pos =
    t_fix ()
    (fn () =>
	(fn recurse =>
	    t_compose_l
		[unfold_r pos,
		 t_succeed (unfold_r pos),
		 apply_r pos,
		 apply_r pos,
		 apply_r pos,
		 cases_r pos,
		 t_orelse
		     (true_r pos)
		     (recurse())
		 ]))

(*****************************************
     Tactics handling at-formula
******************************************)
fun array_at_l pos = 
    t_fix ()
    (fn () =>
	(fn recurse =>
	    t_compose_l
		[unfold_l pos,
		 t_succeed (unfold_l pos),
		 apply_l pos,
		 apply_l pos,
		 apply_l pos,
		 cases_l pos,
		 cases_l pos,
		 t_orelse_l
		     [try_eq_l pos,
		      recurse()
		      ]
		 ]))
    
fun array_at_r pos =
    t_fix ()
    (fn () =>
	(fn recurse =>
	    t_compose_l
		[unfold_r pos,
		 t_succeed (unfold_r pos),
		 apply_r pos,
		 apply_r pos,
		 apply_r pos,
		 cases_r pos,
		 cases_r pos,
		 t_orelse
		     (try_eq_r pos)
		     (recurse())
		 ]))

(*****************************************
     Tactics handling set-formula
******************************************)
fun array_set_l pos =
    t_fix ()
    (fn () =>
	(fn recurse =>
	    t_compose_l
		[unfold_l pos,
		 t_succeed (unfold_l pos),
		 apply_l pos,
		 apply_l pos,
		 apply_l pos,
		 apply_l pos,
		 cases_l pos,
		 cases_l pos,
		 and_l pos,
		 eq_subst pos,
		 t_orelse_l
		     [try_eq_l pos,
		      recurse()
		      ]
		 ]))
fun array_set_r pos =
    t_fix ()
    (fn () =>
	(fn recurse =>
	    t_compose_l
		[unfold_r pos,
		 t_succeed (unfold_r pos),
		 apply_r pos,
		 apply_r pos,
		 apply_r pos,
		 apply_r pos,
		 cases_r pos,
		 cases_r pos,
		 t_compose
		     (and_r pos)
		     [try_eq_r pos,
		      (t_orelse
			  (exteq_r pos)
			  (recurse()))
		      ]
		 ]))

(*****************************************
     Tactics handling size-formula
******************************************)
fun array_size_l pos =
    t_fix ()
    (fn () =>
	(fn recurse =>
	    t_compose_l
		[unfold_l pos,
		 t_succeed (unfold_l pos),
		 apply_l pos,
		 apply_l pos,
		 cases_l pos,
		 handleifeq_l pos
			    [try_eq_l pos
			     ]
			    [cases_l pos,
			     recurse()
			     ]
		 ]))
fun array_size_r pos =
    t_fix ()
    (fn () =>
	(fn recurse =>
	    t_compose_l
		[unfold_r pos,
		 t_succeed (unfold_r pos),
		 apply_r pos,
		 apply_r pos,
		 cases_r pos,
		 handleifeq_r pos
			    [try_eq_r pos
			     ]
			    [cases_r pos,
			     recurse()
			     ]
		 ]))

(*****************************************
     Tactics handling extend-formula
******************************************)
fun array_extend_l pos = 
    t_fix ()
    (fn () =>
	(fn recurse =>
	    t_compose_l
		[unfold_l pos,
		 t_succeed (unfold_l pos),
		 apply_l pos,
		 apply_l pos,
		 apply_l pos,
		 cases_l pos,
		 t_orelse_l
		     [eq_l pos,
		      eq_subst pos,
		      t_compose2
			  (cases_l pos)
			  (recurse())
		      ]
		 ]))
    
fun array_extend_r pos =
    t_fix ()
    (fn () =>
	(fn recurse =>
	    t_compose_l
		[unfold_r pos,
		 t_succeed (unfold_r pos),
		 apply_r pos,
		 apply_r pos,
		 apply_r pos,
		 cases_r pos,
		 t_orelse
		     (exteq_r pos)
		     (t_compose2
			  (cases_r pos)
			  (recurse()))
		 ]))

(*****************************************
     Tactics handling remove-formula
******************************************)
fun array_remove_l pos = 
    t_fix ()
    (fn () =>
	(fn recurse =>
	    t_compose_l
		[unfold_l pos,
		 t_succeed (unfold_l pos),
		 apply_l pos,
		 apply_l pos,
		 apply_l pos,
		 cases_l pos,
		 t_orelse_l
		     [try_eq_l pos,
		      t_compose2
			  (cases_l pos)
			  (recurse())
		      ]
		 ]))
    
fun array_remove_r pos =
    t_fix ()
    (fn () =>
	(fn recurse =>
	    t_compose_l
		[unfold_r pos,
		 t_succeed (unfold_r pos),
		 apply_r pos,
		 apply_r pos,
		 apply_r pos,
		 cases_r pos,
		 t_orelse
		     (try_eq_r pos)
		     (t_compose2
			  (cases_r pos)
			  (recurse()))
		 ]))

(*****************************************
     Tactics handling contains-formula
******************************************)
fun array_contains_l pos =
    t_fix ()
    (fn () =>
	(fn recurse =>
	    t_compose_l
		[unfold_l pos,
		 t_succeed (unfold_l pos),
		 apply_l pos,
		 apply_l pos,
		 apply_l pos,
		 cases_l pos,
		 cases_l pos,
		 (t_compose
		      (if_then_else_l pos)
		      [(t_compose2
			    (apply_to_end_l try_eq_l)
			    (try_eq_l pos)),
		       t_compose2
			   (apply_to_end_r try_eq_r)
			   (recurse())
			   ])
		 ]))

fun array_contains_r pos =
    t_fix ()
    (fn () =>
	(fn recurse =>
	    t_compose_l
		[unfold_r pos,
		 t_succeed (unfold_r pos),
		 apply_r pos,
		 apply_r pos,
		 apply_r pos,
		 cases_r pos,
		 cases_r pos,
		 (t_compose
		      (if_then_else_r pos)
		      [(t_compose2
			    (apply_to_end_l try_eq_l)
			    (try_eq_r pos)),
		       t_compose2
			   (apply_to_end_r try_eq_r)
			   (recurse())
			   ])
		     ]))


(*****************************************
     Tactics handling reverse-formula
******************************************)
fun array_reverse_l pos = 
    t_compose_l
	[unfold_l pos,
	 apply_l pos,
	 apply_l pos,
	 (t_fix ()
		(fn () =>
		    (fn recurse =>
			t_compose_l
			    [t_succeed (unfold_l pos),
			     apply_l pos,
			     apply_l pos,
			     apply_l pos,
			     cases_l pos,
			     (t_orelse_l
				  [eq_l pos,
				   eq_subst pos,
				   (t_compose_l
					[exists_l pos,
					 and_l pos,
					 eq_subst pos,
					 recurse()
					 ])
				   ])
			     ])))
	 ]
fun array_reverse_r pos =
    t_compose_l
	[unfold_r pos,
	 apply_r pos,
	 apply_r pos,
	 (t_fix ()
		(fn () =>
		    (fn recurse =>
			t_compose_l
			    [t_succeed (unfold_r pos),
			     apply_r pos,
			     apply_r pos,
			     apply_r pos,
			     cases_r pos,
			     (t_orelse_l
				  [exteq_r pos,
				   (t_compose_l
					[exists_r NONE pos,
					 t_compose
					     (and_r pos)
					     [exteq_r pos,
					      recurse()]
					     ])
				   ])
			     ])))
	 ]
		
(*****************************************
     Tactics handling set_from-formula
******************************************)
fun array_set_from_l pos = 
    (t_fix ()
     (fn () =>
	 (fn recurse =>
	     t_compose_l
	     [unfold_l pos,
	      t_succeed (unfold_l pos),
	      apply_l pos,
	      apply_l pos,
	      apply_l pos,
	      apply_l pos,
	      cases_l pos,
	      (t_bool
		   (isSetFromL pos)
		   (recurse())
		   (t_fix ()
			  (fn () =>
			      (fn recurse2 =>
				  t_compose_l
				      [t_succeed (unfold_l pos),
				       apply_l pos,
				       apply_l pos,
				       apply_l pos,
				       cases_l pos,
				       (t_orelse_l
					    [eq_l pos,
					     eq_subst pos,
					     (t_compose_l
						  [cases_l pos,
						   recurse2()
						   ])
					     ])
				       ]))))
	      ])))

fun array_set_from_r pos = 
    (t_fix ()
     (fn () =>
	 (fn recurse =>
	     t_compose_l
	     [unfold_r pos,
	      t_succeed (unfold_r pos),
	      apply_r pos,
	      apply_r pos,
	      apply_r pos,
	      apply_r pos,
	      cases_r pos,
	      (t_bool
		   (isSetFromR pos)
		   (recurse())
		   (t_fix ()
			  (fn () =>
			      (fn recurse2 =>
				  t_compose_l
				      [t_succeed (unfold_r pos),
				       apply_r pos,
				       apply_r pos,
				       apply_r pos,
				       cases_r pos,
				       (t_orelse_l
					    [exteq_r pos,
					     (t_compose_l
						  [cases_r pos,
						   recurse2()
						   ])
					     ])
				       ]))))
	      ])))
		
(*****************************************
     Tactics handling split-formula
******************************************)
fun array_split_l pos = 
    t_fix ()
    (fn () =>
	(fn recurse =>
	    t_compose_l
		[unfold_l pos,
		 t_succeed (unfold_l pos),
		 apply_l pos,
		 apply_l pos,
		 apply_l pos,
		 apply_l pos,
		 cases_l pos,
		 t_orelse_l
		     [eq_l pos,
		      eq_subst pos,
		      recurse()
		      ]
		 ]))
    fun array_split_r pos =
    t_fix ()
    (fn () =>
	(fn recurse =>
	    t_compose_l
		[unfold_r pos,
		 t_succeed (unfold_r pos),
		 apply_r pos,
		 apply_r pos,
		 apply_r pos,
		 apply_r pos,
		 cases_r pos,
		 t_orelse
		     (exteq_r pos)
		     (recurse())
		 ]))

(*****************************************
     Tactics handling split-formula
******************************************)
fun array_split_rev_l pos =
    t_compose_l
    [t_succeed (unfold_l pos),
     apply_l pos,
     apply_l pos,
     apply_l pos,
     apply_l pos,
     (t_fix ()
	   (fn () =>
	       (fn recurse =>
		   t_compose_l
		       [unfold_l pos,
			t_succeed (unfold_l pos),
			apply_l pos,
			apply_l pos,
			apply_l pos,
			apply_l pos,
			apply_l pos,
			cases_l pos,
			t_orelse
			    (t_compose_l
				 [and_l pos,
				  try_eq_l pos,
				  try_eq_l pos
				  ])
			    (recurse())
			])))
     ]
    fun array_split_rev_r pos =
	t_compose_l
	    [t_succeed (unfold_r pos),
	     apply_r pos,
	     apply_r pos,
	     apply_r pos,
	     apply_r pos,
	     (t_fix ()
		    (fn () =>
			(fn recurse =>
			    t_compose_l
				[unfold_r pos,
				 t_succeed (unfold_r pos),
				 apply_r pos,
				 apply_r pos,
				 apply_r pos,
				 apply_r pos,
				 apply_r pos,
				 cases_r pos,
				 t_orelse
				     (t_compose2
					  (and_r pos)
					  (try_eq_r pos)
				      )
				     (recurse())
				     ])))
	     ]
	    
(*****************************************
     Commands to ease the use of the
     above tactics in the interactive 
     assistant.
******************************************)
fun Array_at_l        pos = by(array_at_l        pos)
fun Array_at_r        pos = by(array_at_r        pos)
fun Array_set_l       pos = by(array_set_l       pos)
fun Array_set_r       pos = by(array_set_r       pos)
fun Array_size_r      pos = by(array_size_r      pos)
fun Array_size_l      pos = by(array_size_l      pos)
fun Array_split_r     pos = by(array_split_r     pos)
fun Array_split_l     pos = by(array_split_l     pos)
fun Array_create_r    pos = by(array_create_r    pos)
fun Array_create_l    pos = by(array_create_l    pos)
fun Array_extend_l    pos = by(array_extend_l    pos)
fun Array_extend_r    pos = by(array_extend_r    pos)
fun Array_remove_l    pos = by(array_remove_l    pos)
fun Array_remove_r    pos = by(array_remove_r    pos)
fun Array_reverse_l   pos = by(array_reverse_l   pos)
fun Array_reverse_r   pos = by(array_reverse_r   pos)
fun Array_contains_l  pos = by(array_contains_l  pos)
fun Array_contains_r  pos = by(array_contains_r  pos)
fun Array_set_from_l  pos = by(array_set_from_l  pos)
fun Array_set_from_r  pos = by(array_set_from_r  pos)
fun Array_split_rev_l pos = by(array_split_rev_l pos)
fun Array_split_rev_r pos = by(array_split_rev_r pos)
