(*
 * Tactics-file for assisting the user in the task of handling 
 * of the JVM.
 *
 * @todo REALLY shold insert check into the t_while' tactics to make
 * sure that they act on the correct formula. Could be a source of unintended 
 * interesting behaviur otherwise.
 *)


(*****************************************
     Utility tactics
******************************************)
(* Tactics for applying formulas to the end of sequents
 * to the left and the right. Copied from the reference
 * manual.
 *)
val apply_to_end_r =
    t_pretactic
    (fn seq => SOME(length(Sequent.get_rhs seq)))
val apply_to_end_l =
    t_pretactic
    (fn seq => SOME(length(Sequent.get_lhs seq)))
fun eqel_l pos =
    t_orelse
	(eq_l pos)
	(eq_flat_elim_l pos)
fun eqel_r pos =
    t_orelse
	(eq_r pos)
	(eq_flat_elim_r pos)
fun   t_cases ((test, act)::rest) telse =
      t_bool test act (t_cases rest telse)
    | t_cases []                  telse = 
      telse
fun get_next_instruction form = 
    let
	fun fst2 (a,_)   = a
	fun snd2 (_,b)   = b
	val second_form  = fst2(Mmc.application_parts(form))
	val third_form   = fst2(Mmc.application_parts(second_form))
	val first_term   = snd2(Mmc.application_parts(third_form))
	val second_term  = Term.deref first_term
	val mt_args_term = snd2(Term.app_terms(second_term))
	val act_cs_term  = hd (tl (Term.get_tuple mt_args_term))
	val cs_top_term  = snd2(Term.app_terms(act_cs_term))
	val inst_term    = hd (Term.get_tuple cs_top_term)
	val term         = inst_term
    in
	term
    end
fun get_next_instruction_l pos seq =
    get_next_instruction (Sequent.get_formula_lhs(seq,pos))
fun get_next_instruction_r pos seq =
    get_next_instruction (Sequent.get_formula_rhs(seq,pos))


fun disp_inst_l pos =
    t_pretactic
	(fn seq => SOME(get_next_instruction_l pos seq))
	(fn term => 
	    (DaVinci.display_term(term);
	     if (Jvm.is_push term) 
	     then (t_skip)
	     else (t_fail)
		  )
	)
fun disp_inst_r pos =
    t_pretactic
	(fn seq => SOME(get_next_instruction_r pos seq))
	(fn term => 
	    (DaVinci.display_term(term);
	     if (Jvm.is_pop term) 
	     then (t_skip)
	     else (t_fail)
		  )
	)

(*****************************************
     Tactics handling a natPlus-formula
******************************************)
fun natplus_l pos =
    t_fix
	0
	(fn add =>
	    (fn recurse =>
		t_compose_l
		    [unfold_l (pos+add),
		     t_succeed (unfold_l (pos+add)),
		     apply_l (pos+add),
		     apply_l (pos+add),
		     apply_l (pos+add),
		     cases_l (pos+add),
		     t_orelse
			 (eq_subst (pos+add))
			 (t_compose_l
			      [exists_l (pos+add),
			       and_l (pos+add),
			       recurse (add+1),
			       t_orelse
				   (eq_subst (pos+add))
				   (eq_flat_elim_l (pos+add))
			       ])
			 ]
		    )
	    )
fun natplus_r pos =
    t_fix
	0
	(fn _ =>
	    (fn recurse =>
		t_compose_l
		    [unfold_r pos,
		     t_succeed (unfold_r pos),
		     apply_r pos,
		     apply_r pos,
		     apply_r pos,
		     cases_r pos,
		     t_orelse_l
			 [exteq_r pos,
			  eq_flat_elim_r pos,
			  (t_compose_l
			       [exists_r NONE pos,
				t_compose
				    (and_r pos)
				    [exteq_r pos,
				     recurse 0
				     ] 
				    ]
			       )
			  ]
			 ]
		    )
	    )


(*****************************************
     Tactics handling a natMinus-formula
******************************************)
fun natminus_l pos =
    t_fix
	0
	(fn add =>
	    (fn recurse =>
		t_compose_l
		    [unfold_l (pos+add),
		     t_succeed (unfold_l (pos+add)),
		     apply_l (pos+add),
		     apply_l (pos+add),
		     apply_l (pos+add),
		     cases_l (pos+add),
		     t_orelse
			 (eq_subst (pos+add))
			 (t_compose_l
			      [exists_l (pos+add),
			       and_l (pos+add),
			       recurse (add+1),
			       (t_orelse
				   (eq_flat_elim_l (pos+add))
				   (t_compose2
					(eq_flat_decomp_l (pos+add))
					(eq_subst (pos+add))
					)
				   )
			       ])
			 ]
		    )
	    )
fun natminus_r pos =
    natplus_r pos


(*****************************************
     Tactics handling a natLess-formula
******************************************)
fun natless_r pos =
    t_fix
	0
	(fn _ =>
	    (fn recurse =>
		t_compose_l
		    [unfold_r pos,
		     t_succeed (unfold_r pos),
		     apply_r pos,
		     apply_r pos,
		     cases_r pos,
		     exists_r NONE pos,
		     t_orelse
			 (exteq_r pos)
			 (t_compose
			      (and_r pos)
			      [exteq_r pos,
			       recurse 0
			       ]
			      )
			 ]
		    )
	    )
fun natless_l pos =
    t_fix
	0
	(fn _ =>
	    (fn recurse =>
		t_compose_l
		    [unfold_l pos,
		     t_succeed (unfold_l pos),
		     apply_l pos,
		     apply_l pos,
		     cases_l pos,
		     exists_l pos,
		     t_orelse
			 (t_compose_l
			      [and_l pos,
			       t_orelse
				   (eq_flat_elim_l pos)
				   (t_compose_l
					[eq_flat_decomp_l pos,
					 eq_subst pos,
					 t_succeed (recurse 0)
					 ])
				   ])
			 (t_orelse
			      (t_compose_l
				   [eq_flat_decomp_l pos,
				    eq_subst pos
				    ]
				   )
			      (eq_flat_elim_l pos)
			      )
			 ]
		    )
	    )

(*****************************************
     Tactics handling a methodTrans-formula
******************************************)
fun methodtrans_l pos =
    let
	(* Functions to collect common parts in the 
	 * handling of different instructions
	 *)
	fun instTransStart pos =
	    t_compose_l
		[apply_l pos,
		 apply_l pos,
		 apply_l pos,
		 cases_l pos,
		 (*Next handles tau. *)
		 t_succeed (apply_to_end_l eq_l)
		 ]
	val ignoreStore =
	    t_compose_l
		[(*Handles ignored store*)
		 apply_to_end_l eq_flat_decomp_l,
		 apply_to_end_l eq_subst
		 ]
	val noArgStack =
	    t_compose_l
		[(* Next five handles no-argument stack. *)
		 apply_to_end_l eq_subst
		 ]
	val oneArgStack =
	    t_compose_l
		[(* Next five handles one-argument stack. *)
		 apply_to_end_l eq_flat_decomp_l,
		 apply_to_end_l eq_subst,
		 apply_to_end_l eq_subst
		 ]
	val twoArgStack =
	    t_compose_l
		[(* Next five handles two-argument stack. *)
		 apply_to_end_l eq_flat_decomp_l,
		 apply_to_end_l eq_flat_decomp_l,
		 apply_to_end_l eq_subst,
		 apply_to_end_l eq_subst,
		 apply_to_end_l eq_subst
		 ]	
	fun codeStackHandler extras =
	    t_compose_l
		([apply_to_end_l eq_flat_decomp_l,
		 apply_to_end_l eq_subst
		 ]     @
		extras @
		[apply_to_end_l eq_subst
		 ])
	val codeStackNoArg       = codeStackHandler [apply_to_end_l eq_l]
	val codeStackArg         = codeStackHandler [apply_to_end_l eq_flat_decomp_l,
						     apply_to_end_l eq_subst]
	val codeStackWrongNumArg = codeStackHandler [apply_to_end_l eq_flat_decomp_l,
						     apply_to_end_l eq_flat_elim_l]
			     
	fun icXXbf compfalse comptrue =
	    t_compose_l
		[instTransStart pos,
		 t_orelse (*The two possibilities. *)
		     (* Handle false comparison. *)
		     (t_compose_l
			  ([(*Decide if this is false-
			    * comparison-part*)
			   not_l pos,
			   ignoreStore,
			   twoArgStack,
			   codeStackArg
			   (* Handle the comparison. *)
			   ] @ compfalse)
			  )
		     (* Handle true comparison. *)
		     (t_compose_l
			  ([(*Decide if this is true-
			    * comparison-part*)
			   and_l pos,
			   twoArgStack,
			   codeStackArg
			   (*,t_print_sequent ("Checking:") (fn ()=>t_skip)*)
			   ] @ comptrue)
			  )
		     ]
	(* Function used in t_cases for determining
	 * the current instruction
	 *)
	fun check cfun =
	    (fn seq => 
		(
		 (*((t_print_sequent "Checking :" (fn ()=>t_skip)) seq);*)
		 cfun (get_next_instruction_l pos seq)
		 ))
    in
    t_fix 0 (fn _ => (fn recurse =>
	t_compose_l
	    [unfold_l pos,
	     t_succeed (unfold_l pos),
	     apply_l pos,
	     apply_l pos,
	     apply_l pos,
	     cases_l pos,
	     apply_to_end_l eq_subst,
	     apply_to_end_l eq_flat_decomp_l,
	     apply_to_end_l eq_subst,
	     apply_to_end_l eq_subst,
	     apply_to_end_l eq_subst,	 
	     cases_l pos,
	     (* split the eventual double-case, and make sure 
	      * that all of the resulting goals are passed 
	      * through. *)
	     (t_succeed (or_l pos)),
	     t_cases
		 [((check Jvm.is_push),
		   t_compose_l
		       [instTransStart pos,
			ignoreStore,
			noArgStack,
			codeStackArg,
			true_l pos
			]
		   ),
		  ((check Jvm.is_pop),
		   t_compose_l
		       [instTransStart pos,
			ignoreStore,
			oneArgStack,
			codeStackNoArg,
			true_l pos
			]
		   ),
		  ((check Jvm.is_dup),
		   t_compose_l
		       [instTransStart pos,
			ignoreStore,
			oneArgStack,
			codeStackNoArg,
			true_l pos
			]
		   ),
		  ((check Jvm.is_swap),
		   t_compose_l
		       [instTransStart pos,
			ignoreStore,
			twoArgStack,
			codeStackNoArg,
			true_l pos
			]
		   ),
		  ((check Jvm.is_store),
		   t_compose_l
		       [instTransStart pos,
			ignoreStore,
			oneArgStack,
			codeStackArg,
			exists_l pos,
			and_l pos,
			array_set_l pos,
			eq_subst pos
			]
		   ),
		  ((check Jvm.is_load),
		   t_compose_l
		       [instTransStart pos,
			ignoreStore,
			noArgStack,
			codeStackArg,
			exists_l pos,
			and_l pos,
			array_at_l pos,
			eq_subst pos
			]
		   ),
		  ((check Jvm.is_iadd),
		   t_compose_l
		       [instTransStart pos,
			ignoreStore,
			twoArgStack,
			codeStackNoArg,
			(* Handle case-formula (plus) *)
			natplus_l pos
			]
		   ),
		  ((check (fn t => (Jvm.is_gotof t) orelse
				   (Jvm.is_gotob t)
				   )),
		   t_compose_l
		       [instTransStart pos,
			noArgStack,
			codeStackArg,
			recurse 0
			]
		       ),
		  ((check (fn t => (Jvm.is_iceqf t) orelse
				   (Jvm.is_iceqb t)
				   )),
		   t_compose_l
		       [instTransStart pos,
			twoArgStack,
			codeStackArg,
			t_compose
			    (if_then_else_l pos)
			    [(t_compose_l
				 [apply_to_end_l eqel_l,
				  recurse 0
				  ]),
			     (t_compose_l
				  [apply_to_end_r eqel_r,
				   and_l pos,
				   eq_subst pos,
				   eq_l pos
				   ])
			     ]
			]
		   ),
		  ((check (fn t => (Jvm.is_icltf t) orelse
				   (Jvm.is_icltb t)
				   )),
		   t_compose_l
		       [instTransStart pos,
			twoArgStack,
			codeStackArg,
			t_compose
			    (if_then_else_l pos)
			    [(t_compose_l
				 [apply_to_end_l natless_l,
				  recurse 0
				  ]),
			     (t_compose_l
				  [(t_orelse
					(apply_to_end_r natless_r)
					(apply_to_end_r weak_r)
					),
				   and_l pos,
				   eq_subst pos,
				   eq_l pos
				   ])
			     ]
			]
		   ),
		  ((check (fn t => (Jvm.is_i_gotof t) orelse
				   (Jvm.is_i_gotob t)
				   )),
		   t_compose_l
		       [instTransStart pos,
			noArgStack,
			(t_orelse
			     (t_compose_l
				  [(* Eh, something strange.... *)
				   (t_orelse
				       (t_compose_l
					    [apply_to_end_l eq_flat_decomp_l,
					     apply_to_end_l eq_flat_decomp_l,
					     apply_to_end_l eq_subst,
					     apply_to_end_l eq_subst,
					     apply_to_end_l eq_flat_decomp_l,
					     apply_to_end_l eq_subst,
					     apply_to_end_l eq_subst])
				       (t_compose_l
					    [apply_to_end_l eq_flat_decomp_l,
					     apply_to_end_l eq_subst,
					     apply_to_end_l eq_flat_decomp_l,
					     apply_to_end_l eq_subst,
					     (* Hack to handle the end of code instructions... *)
					     t_succeed (apply_to_end_l eq_flat_decomp_l),
					     t_succeed (apply_to_end_l eq_subst),
					     t_succeed (apply_to_end_l eq_subst),
					     t_succeed (apply_to_end_l eq_flat_elim_l) (*horrible hack*)
					     ])),
				   (* ....anyways *)
				   t_compose
				       (if_then_else_l pos)
				       [(t_compose_l
					     [apply_to_end_l eqel_l,
					      and_l pos,
					      eq_subst pos,
					      eq_l pos
					      ]),
					(t_compose_l
					     [apply_to_end_r eqel_r,
					      exists_l pos,
					      and_l pos,
					      eq_flat_decomp_l pos,
					      eq_subst pos,
					      recurse 0
					      ])
					]
				       ])
			     (t_orelse
				  (t_compose_l
				       [apply_to_end_l eq_flat_decomp_l,
					apply_to_end_l eq_subst,
					(t_orelse
					     (apply_to_end_l eq_l)
					     (t_compose_l
						  [apply_to_end_l eq_flat_decomp_l,
						   apply_to_end_l eq_flat_elim_l
						   ])),
					apply_to_end_l eq_l,
					and_l pos,
					eq_subst pos,
					eq_l pos
					])
				  (t_compose_l
				       [apply_to_end_l eq_flat_decomp_l,
					apply_to_end_l eq_subst,
					(t_orelse
					     (apply_to_end_l eq_l)
					     (t_compose_l
						  [apply_to_end_l eq_flat_decomp_l,
						   apply_to_end_l eq_flat_elim_l
						   ])),
					apply_to_end_l eqel_l,
					and_l pos,
					eq_subst pos,
					eq_l pos
					])
				  )
			     )
			]
		   )
		  ]
		 t_fail
	     ]))
    end

(* qqq *)
fun methodtrans_r pos =
    let
	(* Functions to collect common parts in the 
	 * handling of different instructions
	 *)
	fun instTransStart pos =
	    t_compose_l
		[apply_r pos,
		 apply_r pos,
		 apply_r pos,
		 cases_r pos
		 ]
	fun decorate_with_iTS ((a, b) :: rest) =
	    (a, (t_compose_l ((instTransStart pos) :: b))) :: 
	    (decorate_with_iTS rest)
	  | decorate_with_iTS []             =
	    []
	fun handleif pos truepart falsepart =
	    t_compose
		(if_then_else_r pos)
		[t_compose_l truepart,
		 t_compose_l falsepart
		 ]
	fun icXXbf compfalse comptrue =
	    [
	     ]
	(* Function used in t_cases for determining
	 * the current instruction
	 *)
	fun check cfun =
	    (fn seq => 
		(
		 (*((t_print_sequent "Checking :" (fn ()=>t_skip)) seq);*)
		 cfun (get_next_instruction_r pos seq)
		 ))
    in
    t_fix 0 (fn _ => (fn recurse =>
	t_compose_l
	    [unfold_r pos,
	     t_succeed (unfold_r pos),
	     apply_r pos,
	     apply_r pos,
	     apply_r pos,
	     cases_r pos,
	     cases_r pos,
	     t_cases
		 (decorate_with_iTS
		      ([((check Jvm.is_push),
			 [true_r pos
			  ]
			 ),
			((check Jvm.is_pop),
			 [true_r pos
			  ]
			 ),
			((check Jvm.is_dup),
			 [true_r pos
			  ]
			 ),
			((check Jvm.is_swap),
			 [true_r pos
			  ]
			 ),
			((check Jvm.is_store),
			 [exists_r NONE pos,
			  t_compose
			      (and_r pos)
			      [array_set_r pos,
			       exteq_r pos]
			  ]
			 ),
			((check Jvm.is_load),
			 [exists_r NONE pos,
			  t_compose
			      (and_r pos)
			      [array_at_r pos,
			       exteq_r pos]
			  ]
			 ),
			((check Jvm.is_iadd),
			 [(* Handle case-formula (plus) *)
			  natplus_r pos
			  ]
			 ),
			((check (fn t => (Jvm.is_gotof t) orelse
					 (Jvm.is_gotob t)
					 )),
			 [recurse 0
			  ]
			 ),
			((check (fn t => (Jvm.is_iceqf t) orelse
					 (Jvm.is_iceqb t)
					 )),
			 [(handleif pos
				    [(t_orelse
					  (apply_to_end_l eq_l)
					  (apply_to_end_l eq_flat_elim_l)),
				     recurse 0
				     ]
				    [(t_orelse
					  (apply_to_end_r eq_r)
					  (apply_to_end_r eq_flat_elim_r)),
				     (t_compose
					  (and_r pos)
					  [exteq_r pos,
					   exteq_r pos
					   ])
				     ]
				    )
			  ]
			 ),
			((check (fn t => (Jvm.is_icltf t) orelse
					 (Jvm.is_icltb t)
					 )),
			 [(handleif pos
				   [apply_to_end_l natless_l,
				    recurse 0
				    ]
				   [apply_to_end_r natless_r,
				    exteq_r pos
				    ]
				   )
			  ]
			 ),
			((check (fn t => (Jvm.is_i_gotof t) orelse
					 (Jvm.is_i_gotob t)
					 )),
			 [(t_orelse
			       (handleif pos
					 (*True, i.e. stop shifting insts. *)
					 [(t_orelse
					       (apply_to_end_l eq_l)
					       (apply_to_end_l eq_flat_elim_l)),
					  (t_compose
					       (and_r pos)
					       [exteq_r pos,
						exteq_r pos
						])
					  ]
					 (*False, shift another inst and continue. *)
					 [(t_orelse
					       (apply_to_end_r eq_r)
					       (apply_to_end_r eq_flat_elim_r)),
					  exists_r NONE pos,
					  (t_compose
					       (and_r pos)
					       [exteq_r pos,
						recurse 0
						])
					  ])
			       (* Handle boundary-condition... *)
			       (t_compose
				    (and_r pos)
				    [exteq_r pos,
				     exteq_r pos
				     ])
			       )
			  ]
			 )
			])
		      )
		 t_skip
		 ]))
    end

(*****************************************
     Tactics handling a transRel-formula
******************************************)
fun transrel_l  pos =
    t_compose_l
	[unfold_l pos,
	 apply_l pos,
	 apply_l pos,
	 apply_l pos,
	 cases_l pos(*,
	 apply_to_end_l eq_l,
	 apply_to_end_l eq_flat_decomp_l,
	 apply_to_end_l eq_subst,
	 methodtrans_l pos*)
	 ]

(* Probably horribly inefficient, but it may be
 * usefull for small examples.
 *)
fun transrel_l' pos =
    t_compose_l
	[transrel_l pos,
	 (copydischarge1();
	  t_skip)
	 ]

fun transrel_r  pos =
    t_compose_l
	[unfold_r pos,
	 apply_r pos,
	 apply_r pos,
	 apply_r pos,
	 cases_r pos,
	 methodtrans_r pos
	 ]

(* Probably horribly inefficient, but it may be
 * usefull for small examples.
 *)
fun transrel_r' pos =
    t_compose_l
	[transrel_r pos,
	 (copydischarge1();
	  t_skip)
	 ]

(*****************************************
     Commands to ease the use of the
     above tactics in the interactive 
     assistant.
******************************************)
fun Natless_l      pos = by(natless_l   pos)
fun Natless_r      pos = by(natless_r   pos)
fun Natplus_l      pos = by(natplus_l   pos)
fun Natplus_r      pos = by(natplus_r   pos)
fun Natminus_l     pos = by(natminus_l  pos)
fun Natminus_r     pos = by(natminus_r  pos)
fun Transrel_l     pos = by(transrel_l  pos)
fun Transrel_l'    pos = by(transrel_l  pos)
fun Transrel_r     pos = by(transrel_r  pos)
fun Transrel_r'    pos = by(transrel_r' pos)
