(*
 * This file contains some general tactics for simplifying formulae.
 *)


fun   apply_arg (head :: tail) arg =
      (head arg) :: (apply_arg tail arg)
    | apply_arg []             arg =
      []

(*****************************************
    Simplification tacticts.
*****************************************)
fun simplify_tacs_r tacs =
    t_rhs
	(fn pos =>
	    (t_orelse_l
		 ([cases_r pos,
		   if_then_else_r pos
		   ]
		  @
		  (apply_arg tacs pos)
		  @
		  [eq_r pos,
		   true_r pos,
		   false_r pos,
		   not_r pos,
		   or_r pos,
		   and_r pos,
		   eq_flat_decomp_r pos,
		   eq_flat_elim_r pos,
		   forall_r pos,
		   implication_r pos,
		   box_r pos,
		   occurs_check_r pos
		   ]))
	    )

fun simplify_tacs_l tacs =
    t_lhs
	(fn pos =>
	    (t_orelse_l
		 ([cases_l pos,
		   if_then_else_l pos
		   ]
		  @
		  (apply_arg tacs pos)
		  @
		  [eq_l pos,
		   true_l pos,
		   false_l pos,
		   not_l pos,
		   and_l pos,
		   or_l pos,
		   eq_flat_decomp_l pos,
		   eq_flat_elim_l pos,
		   exists_l pos,
		   implication_l pos,
		   diamond_l pos,
		   occurs_check_l pos
		   ]))
	    )
	
fun simplify_tacs ltacs rtacs =
    t_while' 
	(t_orelse_l
	     [id,
	      simplify_tacs_l ltacs,
	      simplify_tacs_r rtacs
	     ]
	 )

val simplify = simplify_tacs [] []
val simplify_l = simplify_tacs_l []
val simplify_r = simplify_tacs_r []


(*************************************************
  Unfolds and applies as much as possible on pos.
 NOTE: Should be reworked to unfold only X times. (X = 2?)
**************************************************)
fun safe_to_unfold_apply form =
    Mmc.is_lfp(form) orelse
    Mmc.is_gfp(form) orelse
    Mmc.is_fpa(form) orelse
    (Mmc.is_application(form) andalso
     let
	 val (form', _) = Mmc.application_parts(form)
     in
	 not (Mmc.is_box(form') orelse
	      Mmc.is_diamond(form'))
     end)
    
fun unfapp_r pos =
    t_while'
	(t_bool 
	     (fn seq => 
		 let 
		     val form = Sequent.get_formula_rhs(seq,pos)
		 in
		     safe_to_unfold_apply form
		 end)
	     (t_orelse_l
		  [apply_r pos,
		   unfold_r pos,
		   unfold_app_r pos,
		   opt_unfold_app_r NONE pos
		   ]
		  )
	     t_fail
	     )

fun unfapp_l pos =
    t_while'
	(t_bool 
	     (fn seq => 
		 let 
		     val form = Sequent.get_formula_lhs(seq,pos)
		 in
		     safe_to_unfold_apply form
		 end)
	     (t_orelse_l
		  [unfold_l pos,
		   unfold_app_l pos,
		   opt_unfold_app_l NONE pos,
		   apply_l pos
		   ]
		  )
	     t_fail
	     )



(*****************************************
     Commands to ease the use of the
     above tactics in the interactive 
     assistant.
******************************************)
fun Unfapp_r        pos         = by(unfapp_r               pos )
fun Unfapp_l        pos         = by(unfapp_l               pos )
fun Simplify_tacs_r tacs        = by(simplify_tacs_r        tacs)
fun Simplify_tacs_l tacs        = by(simplify_tacs_l        tacs)
fun Simplify_tacs   ltacs rtacs = by(simplify_tacs   ltacs rtacs)
fun Simplify_r      ()          = by(simplify_r                 )
fun Simplify_l      ()          = by(simplify_l                 )
fun Simplify        ()          = by(simplify                   )
