(*
 * Tactics-file for assisting the user in the task of handling 
 * of the JVM.
 *
 * @todo REALLY shold insert check into the t_while' tactics to make
 * sure that they act on the correct formula. Could be a source of unintended 
 * interesting behaviur otherwise.
 *)


(*****************************************
     Utility tactics
******************************************)
(* Tactics for applying formulas to the end of sequents
 * to the left and the right. Copied from the reference
 * manual.
 *)
val apply_to_end_r =
    t_pretactic
    (fn seq => SOME(length(Sequent.get_rhs seq)))
val apply_to_end_l =
    t_pretactic
    (fn seq => SOME(length(Sequent.get_lhs seq)))
fun eqel_l pos =
    t_orelse
	(eq_l pos)
	(eq_flat_elim_l pos)
fun eqel_r pos =
    t_orelse
	(eq_r pos)
	(eq_flat_elim_r pos)

fun   t_cases ((test, act)::rest) telse =
      t_bool test act (t_cases rest telse)
    | t_cases []                  telse = 
      telse

fun   t_rep 0 tac =
      t_skip
    | t_rep i tac =
      t_compose2 tac (t_rep (i-1) tac)

fun get_next_instruction form = 
    let
	fun fst2 (a,_)    = a
	fun snd2 (_,b)    = b
	val second_form   = fst2(Mmc.application_parts(form))
	val third_form    = fst2(Mmc.application_parts(second_form))
	val first_term    = fst2(Mmc.application_parts(third_form))
	val typ           = snd2(Mmc.application_parts(first_term))
	val term          = Term.deref typ
    in
	term
    end
fun get_next_instruction_l pos seq =
    get_next_instruction (Sequent.get_formula_lhs(seq,pos))
fun get_next_instruction_r pos seq =
    get_next_instruction (Sequent.get_formula_rhs(seq,pos))


fun disp_inst_l pos =
    t_pretactic
	(fn seq => SOME(get_next_instruction_l pos seq))
	(fn term => 
	    (DaVinci.display_term(term);
	     if (Jvm.is_iceq term) 
	     then (t_skip)
	     else (t_fail)
		  )
	)
fun disp_inst_r pos =
    t_pretactic
	(fn seq => SOME(get_next_instruction_r pos seq))
	(fn term => 
	    (DaVinci.display_term(term);
	     if (Jvm.is_pop term) 
	     then (t_skip)
	     else (t_fail)
		  )
	)

(*****************************************
     Tactics handling a natPlus-formula
******************************************)
fun natplus_l pos =
    t_fix
	0
	(fn add =>
	    (fn recurse =>
		t_compose_l
		    [unfold_l (pos+add),
		     t_succeed (unfold_l (pos+add)),
		     apply_l (pos+add),
		     apply_l (pos+add),
		     apply_l (pos+add),
		     cases_l (pos+add),
		     t_orelse
			 (eq_subst (pos+add))
			 (t_compose_l
			      [exists_l (pos+add),
			       and_l (pos+add),
			       recurse (add+1),
			       t_orelse
				   (eq_subst (pos+add))
				   (eq_flat_elim_l (pos+add))
			       ])
			 ]
		    )
	    )
fun natplus_r pos =
    t_fix
	0
	(fn _ =>
	    (fn recurse =>
		t_compose_l
		    [unfold_r pos,
		     t_succeed (unfold_r pos),
		     apply_r pos,
		     apply_r pos,
		     apply_r pos,
		     cases_r pos,
		     t_orelse_l
			 [exteq_r pos,
			  eq_flat_elim_r pos,
			  (t_compose_l
			       [exists_r NONE pos,
				t_compose
				    (and_r pos)
				    [exteq_r pos,
				     recurse 0
				     ] 
				    ]
			       )
			  ]
			 ]
		    )
	    )


(*****************************************
     Tactics handling a natMinus-formula
******************************************)
fun natminus_l pos =
    t_fix
	0
	(fn add =>
	    (fn recurse =>
		t_compose_l
		    [unfold_l (pos+add),
		     t_succeed (unfold_l (pos+add)),
		     apply_l (pos+add),
		     apply_l (pos+add),
		     apply_l (pos+add),
		     cases_l (pos+add),
		     t_orelse
			 (eq_subst (pos+add))
			 (t_compose_l
			      [exists_l (pos+add),
			       and_l (pos+add),
			       recurse (add+1),
			       (t_orelse
				   (eq_flat_elim_l (pos+add))
				   (t_compose2
					(eq_flat_decomp_l (pos+add))
					(eq_subst (pos+add))
					)
				   )
			       ])
			 ]
		    )
	    )
fun natminus_r pos =
    natplus_r pos


(*****************************************
     Tactics handling a natLess-formula
******************************************)
fun natless_r pos =
    (t_orelse
	 (t_fix
	      0
	      (fn _ =>
		  (fn recurse =>
		      t_compose_l
			  [unfold_r pos,
			   t_succeed (unfold_r pos),
			   apply_r pos,
			   apply_r pos,
			   cases_r pos,
			   exists_r NONE pos,
			   t_orelse
			       (exteq_r pos)
			       (t_compose
				    (and_r pos)
				    [exteq_r pos,
				     recurse 0
				     ]
				    )
			       ]
			  )
		  ))
	 (weak_r pos)
	 )
fun natless_l pos =
    t_fix
	0
	(fn _ =>
	    (fn recurse =>
		t_compose_l
		    [unfold_l pos,
		     t_succeed (unfold_l pos),
		     apply_l pos,
		     apply_l pos,
		     cases_l pos,
		     exists_l pos,
		     t_orelse
			 (t_compose_l
			      [and_l pos,
			       t_orelse
				   (eq_flat_elim_l pos)
				   (t_compose_l
					[eq_flat_decomp_l pos,
					 eq_subst pos,
					 t_succeed (recurse 0)
					 ])
				   ])
			 (t_orelse
			      (t_compose_l
				   [eq_flat_decomp_l pos,
				    eq_subst pos
				    ]
				   )
			      (eq_flat_elim_l pos)
			      )
			 ]
		    )
	    )

(*****************************************
     Tactics handling a natLessEq-formula
******************************************)
fun natlesseq_l pos =
    t_fix 
	() 
	(fn _ => 
	    (fn recurse =>
		t_compose_l
		    [unfold_l pos,
		     t_succeed (unfold_l pos),
		     apply_l pos,
		     apply_l pos,
		     cases_l pos,
		     t_orelse
			 (true_l pos)
			 (t_compose_l
			      [exists_l pos,
			       and_l pos,
			       try_eq_l pos,
			       recurse()
			       ])
		     ]
		))

fun natlesseq_r pos =
    t_fix 
	() 
	(fn _ => 
	    (fn recurse =>
		t_compose_l
		    [unfold_r pos,
		     t_succeed (unfold_r pos),
		     apply_r pos,
		     apply_r pos,
		     cases_r pos,
		     t_orelse
			 (true_r pos)
			 (t_compose_l
			      [exists_r NONE pos,
			       (t_compose
				    (and_r pos)
				    [try_eq_r pos,
				     recurse()])
			       ])
		     ]
		))


(*****************************************
     Tactics handling a methodTrans-formula
******************************************)
fun methodtrans_l pos =
    let
	(* Functions to collect common parts in the 
	 * handling of different instructions
	 *)
	fun instTransStart pos =
	    t_compose_l
		[apply_l pos,
		 apply_l pos,
		 apply_l pos,
		 apply_l pos,
		 cases_l pos
		 ]
	fun decorate_with_iTS ((a, b) :: rest) =
	    (a, (t_compose_l ((instTransStart pos) :: b))) :: 
	    (decorate_with_iTS rest)
	  | decorate_with_iTS []             =
	    []
	(* Function used in t_cases for determining
	 * the current instruction
	 *)
	fun check cfun =
	    (fn seq => 
		(
		 (*((t_print_sequent "Checking :" (fn ()=>t_skip)) seq);*)
		 cfun (get_next_instruction_l pos seq)
		 ))
    in
    t_fix 0 (fn _ => (fn recurse =>
	t_compose_l
	    [t_succeed (unfold_l pos),
	     apply_l pos,
	     apply_l pos,
	     apply_l pos,
	     cases_l pos,
	     exists_l pos,
	     and_l pos,
	     array_at_l pos,
	     cases_l pos,
	     t_cases
		 (decorate_with_iTS
		      [((check (fn i => 
				   (Jvm.is_pop i orelse
				    Jvm.is_dup i orelse
				    Jvm.is_nop i orelse
				    Jvm.is_new i orelse
				    Jvm.is_swap i orelse
				    Jvm.is_goto i orelse
				    Jvm.is_push i orelse
				    Jvm.is_return i orelse
				    Jvm.is_vreturn i orelse
				    Jvm.is_getfield i orelse
				    Jvm.is_putfield i orelse
				    Jvm.is_start i orelse
				    Jvm.is_monitorenter i orelse
				    Jvm.is_monitorexit i
				    ))),
			[true_l pos]
			),
		       ((check Jvm.is_store),
			[array_set_l pos]
			),
		       ((check Jvm.is_load),
			[array_at_l pos]
			),
		       ((check Jvm.is_iadd),
			[natplus_l pos]
			),
		       ((check Jvm.is_isub),
			[natminus_l pos]
			),
		       ((check Jvm.is_iceq),
			[(handleifeq_l pos
				    [eq_subst pos
				     ]
				    [eq_subst pos
				     ])
			 ]
			),
		       ((check Jvm.is_iclt),
			[(handleif_l natless_l natless_r pos
				    [eq_subst pos
				     ]
				    [eq_subst pos
				     ])
			 ]
			),
		       ((check Jvm.is_invoke),
			[array_split_rev_l pos]
			)
		       ])
		      t_fail
		      ]))
    end

(* qqq *)
fun methodtrans_r pos =
    let
	(* Functions to collect common parts in the 
	 * handling of different instructions
	 *)
	fun instTransStart pos =
	    t_compose_l
		[apply_r pos,
		 apply_r pos,
		 apply_r pos,
		 apply_r pos,
		 cases_r pos
		 ]
	fun decorate_with_iTS ((a, b) :: rest) =
	    (a, (t_compose_l ((instTransStart pos) :: b))) :: 
	    (decorate_with_iTS rest)
	  | decorate_with_iTS []             =
	    []
	fun handleif pos truepart falsepart =
	    t_compose
		(if_then_else_r pos)
		[t_compose_l truepart,
		 t_compose_l falsepart
		 ]
	(* Function used in t_cases for determining
	 * the current instruction
	 *)
	fun check cfun =
	    (fn seq => 
		(
		 (*((t_print_sequent "Checking :" (fn ()=>t_skip)) seq);*)
		 cfun (get_next_instruction_r pos seq)
		 ))
    in
    t_fix 0 (fn _ => (fn recurse =>
	t_compose_l
	    [t_succeed (unfold_r pos),
	     apply_r pos,
	     apply_r pos,
	     apply_r pos,
	     cases_r pos,
	     exists_r NONE pos,
	     t_compose
		 (and_r pos)
		 [array_at_r pos,
		  (t_compose_l
		      [cases_r pos,
		       t_cases
			   (decorate_with_iTS
				([((check (fn i => 
					      (Jvm.is_pop i orelse
					       Jvm.is_dup i orelse
					       Jvm.is_nop i orelse
					       Jvm.is_new i orelse
					       Jvm.is_swap i orelse
					       Jvm.is_goto i orelse
					       Jvm.is_push i orelse
					       Jvm.is_return i orelse
					       Jvm.is_vreturn i orelse
					       Jvm.is_getfield i orelse
					       Jvm.is_putfield i orelse
					       Jvm.is_start i orelse
					       Jvm.is_monitorenter i orelse
					       Jvm.is_monitorexit i
					       ))),
				   [true_r pos]
				   ),
				  ((check Jvm.is_store),
				   [array_set_r pos]
				   ),
				  ((check Jvm.is_load),
				   [array_at_r pos]
				   ),
				  ((check Jvm.is_iadd),
				   [natplus_r pos]
				   ),
				  ((check Jvm.is_isub),
				   [natminus_r pos]
				   ),
				  ((check Jvm.is_iceq),
				   [(handleifeq_r pos
						  [exteq_r pos
						   ]
						  [exteq_r pos
						   ])
				    ]
				   ),
				  ((check Jvm.is_iclt),
				   [(handleif_r natless_l natless_r pos
						[exteq_r pos
						 ]
						[exteq_r pos
						 ])
				    ]
				   ),
				  ((check Jvm.is_invoke),
				   [array_split_rev_r pos]
				   )
				  ])
				)
			   t_fail
			   ])
		      ]
		 ]))
    end

(*****************************************
     Tactics handling a transRel-formula
******************************************)
fun transrelonethread_l  pos =
    t_compose_l
	[t_succeed (unfold_l pos),
	 apply_l pos,
	 apply_l pos,
	 apply_l pos,
	 apply_l pos,
	 apply_l pos,
	 apply_l pos,
	 exists_l pos,
	 exists_l pos,
	 and_l pos, array_at_l pos,
	 cases_l pos,
	 exists_l pos,
	 exists_l pos,
	 and_l pos,
	 methodtrans_l pos,
	 cases_l pos,
	 (t_orelse_l
	      [(* tau()-action *)
	       (t_compose_l
		    [and_l pos,
		     try_eq_l pos,
		     array_set_l pos
		     ]),
	       (* result()-action *)
	       (t_compose_l
		    [and_l pos,
		     try_eq_l pos,
		     (handleifeq_l pos
				   [array_remove_l pos]
				   [array_set_l pos])
		     ]),
	       (*
	       eq_subst pos,       (* try_eq_l? *)*)
	       (* vresult()-action *)
	       (t_compose_l
		    [cases_l pos,
		     and_l pos, try_eq_l pos,
		     and_l pos, try_eq_l pos,
		     array_set_l pos
		     ]),
	       (* call(M,A)-action *)
	       (t_compose_l
		    [exists_l pos, (*find method *)
		     exists_l pos,
		     exists_l pos,
		     cases_l pos,
		     and_l pos, array_contains_l pos,
		     exists_l pos, (* find locals *)
		     exists_l pos,
		     and_l pos, array_create_l pos,
		     and_l pos, array_set_from_l pos,
		     and_l pos, try_eq_l pos,
		     and_l pos, try_eq_l pos,
		     array_set_l pos
		     ]),
	       (* get(R,I,V) *)
	       (t_compose_l
		    [exists_l pos,
		     exists_l pos,
		     exists_l pos,
		     exists_l pos,
		     and_l pos, array_at_l pos,
		     and_l pos, array_at_l pos,
		     and_l pos, try_eq_l pos,
		     array_set_l pos
		     ]),
	       (* put(R,I,V) *)
	       (t_compose_l
		    [exists_l pos,
		     exists_l pos,
		     exists_l pos,
		     exists_l pos,
		     exists_l pos,
		     and_l pos, array_at_l pos,
		     and_l pos, array_set_l pos,
		     and_l pos, array_set_l pos,
		     array_set_l pos
		     ]),
	       (* allocate(C, R) *)
	       (t_compose_l
		    [exists_l pos,
		     exists_l pos,
		     cases_l pos,
		     and_l pos, array_contains_l pos,
		     and_l pos, array_create_l pos,
		     and_l pos, array_extend_l pos,
		     and_l pos, array_size_l pos,
		     array_set_l pos
		     ]),
	       (* spawn(Ref, RunId) *)
	       (t_compose_l
		    [t_rep 6 (exists_l pos),
		     cases_l pos,
		     and_l pos, array_contains_l pos,
		     and_l pos, array_create_l pos,
		     and_l pos, array_set_l pos,
		     and_l pos, try_eq_l pos,
		     and_l pos, array_extend_l pos,
		     and_l pos, try_eq_l pos,
		     array_set_l pos
		     ]),
	       (* enter(Ref) *)
	       (t_compose_l
		    [t_rep 4 (exists_l pos),
		     and_l pos, array_at_l pos,
		     handleifeq_l pos
				  [exists_l pos,
				   and_l pos, array_extend_l pos,
				   and_l pos, array_set_l pos,
				   array_remove_l pos
				   ]
				  [and_l pos, array_set_l pos,
				   array_set_l pos
				   ]	  
		     ]),
	       (* *)
	       (t_compose_l
		    [t_rep 3 (exists_l pos),
		     and_l pos, array_at_l pos,
		     cases_l pos,
		     t_orelse
			 (t_compose_l
			      [and_l pos, array_set_l pos,
			       array_set_l pos
			       ])
			 (t_compose_l
			      [exists_l pos,
			       and_l pos, array_set_l pos,
			       and_l pos, array_extend_l pos,
			       array_set_l pos
			       ])
		     ])
	       ])
	 ]

fun transrel_l pos =
    t_compose_l
	[unfold_l pos,
	 apply_l pos,
	 apply_l pos,
	 apply_l pos,
	 cases_l pos,
	 exists_l pos,
	 and_l pos, array_size_l pos,
	 (t_fix 
	      () 
	      (fn _ => 
		  (fn recurse =>
		      t_compose_l
			  [unfold_l pos,
			   t_succeed (unfold_l pos),
			   apply_l pos,
			   apply_l pos,
			   apply_l pos,
			   cases_l pos,
			   (t_compose
				(or_l pos)
				[(t_compose_l
				      [exists_l pos,
				       exists_l pos,
				       and_l pos, 
				       transrelonethread_l pos,
				       try_eq_l pos]),
				 (t_compose_l
				       [cases_l pos,
					(t_orelse
					    (false_l pos)
					    (recurse()))
					])
				 ])
			   ]
	  )))
	 ]

(* Probably horribly inefficient, but it may be
 * usefull for medium-sized examples.
 *)
fun Transrel_l' pos =
(* Does not seem to work...
    t_compose_l
	[transrel_l pos,
	 (copydischarge1();
	  t_skip)
	 ]
*)
    (by(transrel_l pos);
     copydischarge1()
     )

fun andsplit t pos =
    t_compose
	(and_r pos)
	[t pos,
	 t_skip
	 ]

fun transrelonethread_r  pos =
    t_compose_l
	[t_succeed (unfold_r pos),
	 apply_r pos,
	 apply_r pos,
	 apply_r pos,
	 apply_r pos,
	 apply_r pos,
	 apply_r pos,
	 exists_r NONE pos,
	 exists_r NONE pos,
	 andsplit array_at_r pos,
	 cases_r pos,
	 exists_r NONE pos,
	 exists_r NONE pos,	 
	 andsplit methodtrans_r pos,
	 cases_r pos,
	 (t_orelse_l
	      [(* tau()-action *)
	       (t_compose_l
		   [andsplit try_eq_r pos,
		    array_set_r pos
		   ]),
	       (* result()-action *)
	       (t_compose_l
		   [andsplit try_eq_r pos,
		    (handleifeq_r pos
				  [array_remove_r pos]
				  [array_set_r pos])
		    ]),
	       (* vresult()-action *)
	       (t_compose_l
		    [cases_r pos,
		     andsplit try_eq_r pos,
		     andsplit try_eq_r pos,
		     array_set_r pos
		     ]),
	       (* call(M,A)-action *)
	       (t_compose_l
		    [exists_r NONE pos,
		     exists_r NONE pos,
		     exists_r NONE pos,
                     cases_r pos,
		     andsplit array_contains_r pos,
		     exists_r NONE pos,
		     exists_r NONE pos,
		     andsplit array_create_r pos,
		     andsplit array_set_from_r pos,
		     andsplit try_eq_r pos,
		     andsplit try_eq_r pos,
		     array_set_r pos
		     ]),
	       (* get(R,I,V) *)
	       (t_compose_l
		    [exists_r NONE pos,
		     exists_r NONE pos,
		     exists_r NONE pos,
		     exists_r NONE pos,
		     andsplit array_at_r pos,
		     andsplit array_at_r pos,
		     andsplit try_eq_r pos,
		     array_set_r pos
		     ]),
	       (* put(R,I,V) *)
	       (t_compose_l
		    [exists_r NONE pos,
		     exists_r NONE pos,
		     exists_r NONE pos,
		     exists_r NONE pos,
		     exists_r NONE pos,
		     andsplit array_at_r pos,
		     andsplit array_set_r pos,
		     andsplit array_set_r pos,
		     array_set_r pos
		     ]),
	       (* allocate(C,R) *)
	       (t_compose_l
		    [exists_r NONE pos,
		     exists_r NONE pos,
                     cases_r pos,
		     andsplit array_contains_r pos,
		     andsplit array_create_r pos,
		     andsplit array_extend_r pos,
		     andsplit array_size_r pos,
		     array_set_r pos
		     ]),
	       (* spawn(Ref, MId) *)
	       (t_compose_l
		    [t_rep 6 (exists_r NONE pos),
                     cases_r pos,
		     andsplit array_contains_r pos, 
		     andsplit array_create_r pos, 
		     andsplit array_set_r pos, 
		     andsplit try_eq_r pos,
		     andsplit array_extend_r pos, 
		     andsplit try_eq_r pos,
		     array_set_r pos
		     ]),
	       (* enter(Ref) *)
	       (t_compose_l
		    [t_rep 4 (exists_r NONE pos),
		     andsplit array_at_r pos,
		     handleifeq_r pos
			 [exists_r NONE pos,
			  andsplit array_extend_r pos,
			  andsplit array_set_r pos,
			  array_remove_r pos
			  ]
			 [andsplit array_set_r pos,
			  array_set_r pos
			  ]
		     ]),
	       (* exit(Ref) *)
	       (t_compose_l
		    [t_rep 3 (exists_r NONE pos),
		     andsplit array_at_r pos,
		     cases_r pos,
		     t_orelse
			 (t_compose_l
			      [andsplit array_set_r pos,
			       array_set_r pos
			       ])
			 (t_compose_l
			      [exists_r NONE pos,
			       andsplit array_set_r pos,
			       andsplit array_extend_r pos,
			       array_set_r pos
			       ])
		     ])
	       ])
	 ]


fun transrelimpl_r pos =
    t_compose_l
	[unfold_r pos,
	 t_succeed (unfold_r pos),
	 apply_r pos,
	 apply_r pos,
	 apply_r pos,
	 cases_r pos,
	 or_r pos,
	 (* Handle this thread. *)
	 exists_r NONE pos,
	 exists_r NONE pos,
	 andsplit transrelonethread_r pos,
	 move_to_end_r pos,
	 (* Handle possible next thread*)
	 cases_r pos,
	 (t_orelse
	      (false_r pos)
	      (t_skip))
	 ]
    
fun transreltoimpl_r pos = 
    t_compose_l
	[unfapp_r pos,
	 cases_r pos,
	 exists_r NONE pos,
	 andsplit array_size_r pos
	 ]

fun transrel_r pos =
    t_compose2
	(transreltoimpl_r pos)
	(t_while (transrelimpl_r pos))

(* This version should work. Why not, is beyond me...
fun transrel_r pos = 
    t_compose_l
	[unfold_r pos,
	 apply_r pos,
	 apply_r pos,
	 apply_r pos,
	 cases_r pos,
	 exists_r NONE pos,
	 andsplit array_size_r pos,
	 (t_fix 
	      ()
	      (fn _ => 
		  (fn recurse =>
		      t_compose_l
			  [unfold_r pos,
			   t_succeed (unfold_r pos),
			   apply_r pos,
			   apply_r pos,
			   apply_r pos,
			   cases_r pos,
			   or_r pos,
			   (* Handle this thread. *)
			   exists_r NONE pos,
			   exists_r NONE pos,
			   andsplit transrelonethread_r pos,
			   move_to_end_r pos,
			   (*try_eq_r pos,*)
			   (* Handle possible next thread*)
			   cases_r pos,
			   (t_orelse
				(false_r pos)
				(recurse()))
			   ]
	  )))
	 ]
*)
(*old version, keeps things in order and iterates through possiblities.
fun transrel_r pos = 
    t_compose_l
	[unfold_r pos,
	 apply_r pos,
	 apply_r pos,
	 apply_r pos,
	 cases_r pos,
	 exists_r NONE pos,
	 andsplit array_size_r pos,
	 (t_fix 
	      0 
	      (fn add => 
		  (fn recurse =>
		      t_compose_l
			  [unfold_r (pos+add),
			   t_succeed (unfold_r (pos+add)),
			   apply_r (pos+add),
			   apply_r (pos+add),
			   apply_r (pos+add),
			   cases_r (pos+add),
			   or_r (pos+add),
			   (* Handle this thread. *)
			   exists_r NONE (pos+add),
			   exists_r NONE (pos+add),
			   andsplit transrelonethread_r (pos+add),
			   (*try_eq_r (pos+add),*)
			   (* Handle possible next thread*)
			   cases_r (pos+add+1),
			   (t_orelse
				(false_r (pos+add+1))
				(recurse (add+1)))
			   ]
	  )))
	 ]
*)
(* Probably horribly inefficient, but it may be
 * usefull for medium-sized examples.
 *)
fun Transrel_r' pos =
(* This cute construction does not seem to work...
    t_compose_l
	[transrel_r pos,
	 (copydischarge1();
	  t_skip)
	 ]
*)
    (by(transrel_r pos);
     copydischarge1()
     )
    
(*****************************************
     Commands to ease the use of the
     above tactics in the interactive 
     assistant.
******************************************)
fun Natless_l      pos = by(natless_l   pos)
fun Natless_r      pos = by(natless_r   pos)
fun Natplus_l      pos = by(natplus_l   pos)
fun Natplus_r      pos = by(natplus_r   pos)
fun Natminus_l     pos = by(natminus_l  pos)
fun Natminus_r     pos = by(natminus_r  pos)
fun Natlesseq_l    pos = by(natlesseq_l pos)
fun Natlesseq_r    pos = by(natlesseq_r pos)
fun Transrel_l     pos = by(transrel_l  pos)
(*
fun Transrel_l'    pos = by(transrel_l  pos)
*)
fun Transrel_r     pos = by(transrel_r  pos)
(*
fun Transrel_r'    pos = by(transrel_r' pos)
*)
